/*
 * Copyright (c) 2013 Carlo Augusto Grazia, Paolo Valente
 * Universita` di Modena e Reggio Emilia
 * All rights reserved 
 * Initial version by Luigi Rizzo.
 *
 * Testing program for schedulers
 *
 * The framework include a simple controller which, at each
 * iteration, decides whether we can enqueue and/or dequeue.
 * Then the mainloop runs the required number of tests,
 * keeping track of statistics.
 */

#include "dn_test.h"
#include <sys/time.h>
#include <math.h>


struct q_list {
    struct list_head h;
};

struct cfg_s {
    int ac;
    char * const *av;

    const char *name;
    int loops;
    struct timeval time;

    /* running counters */
    uint32_t    _enqueue;
    uint32_t    drop;
    uint32_t    pending;
    uint32_t    dequeue;

    /* generator parameters */
    int th_min, th_max;
    int th_sing; /* 
                  * minimum number of dequeued packets that triggers
                  * a new enqueue phase
                  */
    int maxburst; /* maximum bursts length [pkts] */
    int lmin, lmax; /* packet len */
    int flows;  /* number of flows */
    int flowsets;   /* number of flowsets */
    int wsum;   /* sum of weights of all flows */
    int max_y;  /* max random number in the generation */
    int cur_y, cur_fs;  /* used in generation, between 0 and max_y - 1 */
    const char *fs_config; /* flowset config */
    int can_dequeue;
    uint32_t to_enqueue; /* id of the flow to fill in enqueue */
    int print_stats; /* 
                      * == 1 to print statistics on queue occupancy, mainly for
                      * debugging; 
                      * == 2 or more to print statistics on throughput, T-WFI,
                      * B-WFI, delay.
                      */
    struct mbuf *tosend;    /* packet to send -- also flag to enqueue */

    struct mbuf *freelist;

    struct mbuf *head, *tail;   /* a simple tailq */

    /* scheduler hooks */
    int (*enq)(struct dn_sch_inst *, int main_flow_id);
    int (*deq)(struct dn_sch_inst *);

    void (*finalprint_stats)(struct dn_sch_inst *); /*
                                         * linked to free_sched() funtion of
                                         * double and single container scheduler
                                         */

    /* 
     * The following block of objects should be revisited to remove
     * all the info that have actualy migrated to single- and double-
     * scheduler wrappers.
     */
    /* size of the three fields including sched-specific areas */
    int schk_len;
    int q_len; /* size of a queue including sched-fields */
    int si_len; /* size of a sch_inst including sched-fields */
    char *q;    /* array of flow queues */
        /* use a char* because size is variable */
    struct dn_fsk *fs;  /* array of flowsets */
    struct dn_sch_inst *si;
    struct dn_schk *sched;

    /* generator state */
    int state;      /* 0 = going up, 1: going down */

    /*
     * We keep lists for each backlog level, and always serve
     * the one with shortest backlog. llmask contains a bitmap
     * of lists, and ll are the heads of the lists. The last
     * entry (BACKLOG) contains all entries considered 'full'
     * XXX to optimize things, entry i could contain queues with
     * 2^{i-1}+1 .. 2^i entries.
     */
#define BACKLOG 30 // maximum possible backlog for each flow
    uint32_t    llmask;
    struct list_head ll[BACKLOG + 10];

    int burst_left; /* to implement per-flow bursty arrivals */
};

/* 
 * FI2Q and Q2FI converts from flow_id to dn_queue and back.
 * We cannot easily use pointer arithmetic because it is variable size.
 */
#define FI2Q(c, i)  ((struct dn_queue *)((c)->q + (c)->q_len * (i)))
#define Q2FI(c, q)  (((char *)(q) - (c)->q)/(c)->q_len)

int debug = 0;

static void controller(struct cfg_s *c);

/* decrease the length of the queue */
int
drop(struct cfg_s *c, int flow_id)
{
    struct dn_queue *q;
    int i;

    c->drop++;
    q = FI2Q(c, flow_id);
    i = q->ni.length;/*
                      * Assuming that, when this function is
                      * invoked, q->ni.length is equal to the
                      * index of the flow list minus one in
                      * case the flow is in ll.
                      */

    ND("q %p flow_id %d current length %d", q, flow_id, i);
    if (i < BACKLOG) {
        struct list_head *h = &q->ni.h;
        /*
         * Probably, the next bit-reset should be done more
         * correctly only if the list becomes empty, but the
         * mechanism seems however to work, for the following
         * reason. Before enqueuing a packet for queues at
         * level i+1, all queues at level i must be moved to
         * level i+1. And when the latter operation is
         * completed, the (i+1)-th bit in llmask is certainly
         * set.
         */

        c->llmask |= (1<<(i));

        if (is_in_list(h)) {
            if (list_empty(&c->ll[i+1])) {
                D("failure: list empty %d", i+1);
                exit(1);
            }
            list_del(h);

            /* no more queues at this backlog level */
            if (list_empty(&c->ll[i+1])) 
                c->llmask &= ~(1<<(i+1));
            

        }
        list_add_tail(h, &c->ll[i]);
    }
    return 0;
}

/* enqueue returns NON-NULL when a packet is dropped */
static int
enqueue(struct cfg_s *c, int flow_id)
{
        int failure = 0;
        struct dn_queue *q = FI2Q(c, flow_id);

        if (c->enq)  
                failure = c->enq(c->si, flow_id);
        /*
         * Make sure that each flow is accounted for the number of
         * packets of that flow that are pending
         */
        q->ni.length++;
        return failure; /* default - success */
}

/* dequeue returns flow_id if a packet is available, otherwise -1 */
static int
dequeue(struct cfg_s *c)
{
        int flow_id = 0;
        struct dn_queue *q;
        if (c->deq) 
                flow_id = c->deq(c->si); /* 
                                          * deq() return the flow_id 
                                          * of the packet dequeued
                                          */

        if (flow_id == -1)
                return flow_id;

        q = FI2Q(c, flow_id);
        /*
         * Make sure that each flow is accounted (only) for the number
         * of packets of that flow that are pending
         */
        q->ni.length--;
        return flow_id;
}

static int
mainloop(struct cfg_s *c)
{

    int i;
    int flow_id;
    struct dn_queue *q;

    for (i=0; i < c->loops; i++) {
        /* implement histeresis */
        controller(c);
        ND("post_controller: state %d, can_dequeue %d, to_enqueue %d"
           ,c->state,c->can_dequeue,c->to_enqueue);


        ND("loop %d enq %d send %d rx %d",
            i, c->_enqueue, c->to_enqueue, c->can_dequeue);
        if ( c->to_enqueue != -1 ) {
            flow_id = c->to_enqueue;
            c->_enqueue++;
            q = FI2Q(c, flow_id);
            ND("enqueue flow_id %d, len %d",flow_id, q->ni.length);
            if (enqueue(c, flow_id)) {
                drop(c, flow_id);
                ND("loop %d enqueue fail", i);
            } else {
                c->pending++;

                ND(
                   "enqueue flow_id %d "
                   "pend %d (%d %d)",
                   flow_id,
                   c->pending, c->th_min, c->th_max,
                   );
            }
        }
        if (c->can_dequeue) {
            c->dequeue++;
            c->pending--;
            flow_id = dequeue(c);
            q = FI2Q(c, flow_id);


                ND("dequeue flow_id %d pend %d "
                   "(%d %d)",
                   m->flow_id,
                   c->pending, c->th_min, c->th_max);
                ND("dequeue flow_id %d",flow_id);
                drop(c, flow_id);
                c->drop--;  /* compensate */
            
            if(q->ni.length <= c->th_sing && (rand()%2 == 0)){
                c->state = 0; /* enqueue! */
                ND("FORCED ENQUEUE, state = %d",c->state);
            }

        }
    }
    DX(2, "mainloop ends %d", i);
    return 0;
}

int
dump(struct cfg_s *c)
{
    int i;
    struct dn_queue *q;

    for (i=0; i < c->flows; i++) {
        q = FI2Q(c, i);
        DX(2, "queue %4d tot %10" PRIu64, i, q->ni.tot_bytes);
    }
    DX(2, "done %d loops\n", c->loops);
    return 0;
}

/* interpret a number in human form */
static long
getnum(const char *s, char **next, const char *key)
{
    char *end = NULL;
    long l;

    if (next)   /* default */
        *next = NULL;
    if (s && *s) {
        DX(4, "token is <%s> %s", s, key ? key : "-");
        l = strtol(s, &end, 0);
    } else {
        DX(4, "empty string");
        l = -1;
    }
    if (l < 0) {
        DX(3, "invalid %s for %s", s ? s : "NULL", (key ? key : "") );
        return 0;   // invalid 
    }
    if (!end || !*end)
        return l;
    if (*end == 'n')
        l = -l; /* multiply by n */
    else if (*end == 'K')
        l = l*1000;
    else if (*end == 'M')
        l = l*1000000;
    else if (*end == 'k')
        l = l*1024;
    else if (*end == 'm')
        l = l*1024*1024;
    else if (*end == 'w')
        ;
    else {/* not recognized */
        D("suffix %s for %s, next %p", end, key, next);
        end--;
    }
    end++;
    DX(4, "suffix now %s for %s, next %p", end, key, next);
    if (next && *end) {
        DX(4, "setting next to %s for %s", end, key);
        *next = end;
    }
    return l;
}

/*
 * flowsets are a comma-separated list of
 *     weight:maxlen:flows
 * indicating how many flows are hooked to that fs.
 * Both weight and range can be min-max-steps.
 * In a first pass we just count the number of flowsets and flows,
 * in a second pass we complete the setup.
 */
static void
parse_flowsets(struct cfg_s *c, const char *fs, int pass)
{
    char *s, *cur, *next;
    int n_flows = 0, n_fs = 0, wsum = 0;
    int i, j;
    struct dn_fs *prev = NULL;

    DX(4, "--- pass %d flows %d flowsets %d", pass, c->flows, c->flowsets);
    if (pass == 0)
        c->fs_config = fs;
    s = c->fs_config ? strdup(c->fs_config) : NULL;
    if (s == NULL) {
        if (pass == 0)
            D("no fsconfig");
        return;
    }
    for (next = s; (cur = strsep(&next, ","));) {
        char *p = NULL;
        int w, w_h, w_steps, wi;
        int len, len_h, l_steps, li;
        int flows;

        w = getnum(strsep(&cur, ":"), &p, "weight");
        if (w <= 0)
            w = 1;
        w_h = p ? getnum(p+1, &p, "weight_max") : w;
        w_steps = p ? getnum(p+1, &p, "w_steps") : (w_h == w ?1:2);
        len = getnum(strsep(&cur, ":"), &p, "len");
        if (len <= 0)
            len = 1000;
        len_h = p ? getnum(p+1, &p, "len_max") : len;
        l_steps = p ? getnum(p+1, &p, "l_steps") : (len_h == len ? 1 : 2);
        flows = getnum(strsep(&cur, ":"), NULL, "flows");
        if (flows == 0)
            flows = 1;
        DX(5, "weight %d..%d (%d) len %d..%d (%d) flows %d",
            w, w_h, w_steps, len, len_h, l_steps, flows);
        if (w == 0 || w_h < w || len == 0 || len_h < len ||
                flows == 0) {
            DX(5,"wrong parameters %s", fs);
            return;
        }
        n_flows += flows * w_steps * l_steps;
        for (i = 0; i < w_steps; i++) {
            wi = w + ((w_h - w)* i)/(w_steps == 1 ? 1 : (w_steps-1));
            for (j = 0; j < l_steps; j++, n_fs++) {
                struct dn_fs *fs = &c->fs[n_fs].fs; // tentative
                int x;

                li = len + ((len_h - len)* j)/(l_steps == 1 ? 1 : (l_steps-1));
                x = (wi*2048)/li;
                DX(4, "----- fs %4d weight %4d lmax %4d X %4d flows %d",
                    n_fs, wi, li, x, flows);
                if (pass == 0)
                    continue;
                if (c->fs == NULL || c->flowsets <= n_fs) {
                    D("error in number of flowsets");
                    return;
                }
                wsum += wi * flows;
                fs->par[0] = wi;
                fs->par[1] = li;
                fs->index = n_fs;
                fs->n_flows = flows;
                fs->cur = fs->first_flow = prev==NULL ? 0 : prev->next_flow;
                fs->next_flow = fs->first_flow + fs->n_flows;
                fs->y = x * flows;
                fs->base_y = (prev == NULL) ? 0 : prev->next_y;
                fs->next_y = fs->base_y + fs->y;
                prev = fs;
            }
        }
    }
    c->max_y = prev ? prev->base_y + prev->y : 0;
    c->flows = n_flows;
    c->flowsets = n_fs;
    c->wsum = wsum;
    if (pass == 0)
        return;

    /* now link all flows to their parent flowsets */
    DX(2,"%d flows on %d flowsets max_y %d", c->flows, c->flowsets, c->max_y);
    for (i=0; i < c->flowsets; i++) {
        struct dn_fs *fs = &c->fs[i].fs;
        DX(2, "fs %3d w %5d l %4d flow %5d .. %5d y %6d .. %6d",
            i, fs->par[0], fs->par[1],
            fs->first_flow, fs->next_flow,
            fs->base_y, fs->next_y);
        for (j = fs->first_flow; j < fs->next_flow; j++) {
            struct dn_queue *q = FI2Q(c, j);
            q->fs = &c->fs[i];
        }
    }
}


static int
init(struct cfg_s *c)
{
    int i;
    int ac = c->ac;
    char * const *av = c->av;
    char * const *arg_double = c->av;
    int n_arg_double = c->ac;


    c->si_len = sizeof(struct dn_sch_inst);
    c->q_len = sizeof(struct dn_queue);
    moduledata_t *mod = NULL;
    struct dn_alg *p = NULL;

    c->print_stats = 0;
    c->th_sing = -1;
    c->th_min = 0;
    c->th_max = -20;/* 20 packets per flow */
    c->lmin = c->lmax = 1280;   /* packet len */
    c->flows = 1;
    c->flowsets = 1;
    c->name = "null";
    ac--; av++;
    while (ac > 1) {
        if (!strcmp(*av, "-n")) {
            c->loops = getnum(av[1], NULL, av[0]);
        } else if (!strcmp(*av, "-d")) {
            debug = atoi(av[1]);
#ifdef IPFW
        } else if (!strcmp(*av, "-mode")) {
            extern moduledata_t *_g_dn_double;
            extern moduledata_t *_g_dn_single;
            if (!strcmp(av[1], "single"))
                mod = _g_dn_single;
            else if (!strcmp(av[1], "double"))
                mod = _g_dn_double;
            else
                mod = NULL;
            c->name = mod ? mod->name : "NULL";
            DX(4, "using scheduler %s", c->name);
#endif
        } else if (!strcmp(*av, "-len")) {
            c->lmin = getnum(av[1], NULL, av[0]);
            c->lmax = c->lmin;
            DX(4, "setting max to %d", c->lmax);
        } else if (!strcmp(*av, "-burst")) {
            c->maxburst = getnum(av[1], NULL, av[0]);
            DX(4, "setting max to %d", c->maxburst);
        } else if (!strcmp(*av, "-stats")) {
            c->print_stats = getnum(av[1], NULL, av[0]);
            DX(4, "setting print stats to %d", c->print_stats);
        } else if (!strcmp(*av, "-qmax")) {
            c->th_max = getnum(av[1], NULL, av[0]);
            DX(4, "setting max to %d", c->th_max);
        } else if (!strcmp(*av, "-qmin")) {
            c->th_min = getnum(av[1], NULL, av[0]);
            DX(4, "setting min to %d", c->th_min);
        } else if (!strcmp(*av, "-qsing")) {
            c->th_sing = getnum(av[1], NULL, av[0]);
            DX(4, "setting min for flow to %d", c->th_min);
        } else if (!strcmp(*av, "-flows")) {
            c->flows = getnum(av[1], NULL, av[0]);
            DX(4, "setting flows to %d", c->flows);
        } else if (!strcmp(*av, "-flowsets")) {
            parse_flowsets(c, av[1], 0);
            DX(4, "setting flowsets to %d", c->flowsets);
        } else {
            DX(1,"option %s not recognised, ignore", *av);
        }
        ac -= 2; av += 2;
    }
    if (c->print_stats < 0)
        c->print_stats = 0;
    if (c->th_sing > 30 )
        c->th_sing = -1;
    if (c->maxburst < 0)
        c->maxburst = 0;
    if (c->loops <= 0)
        c->loops = 1;
    if (c->flows <= 0)
        c->flows = 1;
    if (c->flowsets <= 0)
        c->flowsets = 1;
    if (c->lmin <= 0)
        c->lmin = 1;
    if (c->lmax <= 0)
        c->lmax = 1;


    /* multiply by N */
    if (c->th_min < 0)
        c->th_min = c->flows * -c->th_min;
    if (c->th_max < 0)
        c->th_max = c->flows * -c->th_max;
    if (c->th_max <= c->th_min) /* default histeresis of 1 packet */
        c->th_max = c->th_min + 1;

    extern moduledata_t *_g_dn_double;
    extern moduledata_t *_g_dn_single;


    if (mod == NULL){
        printf("no -mode option found: choose between single and double \n");
        exit(1);
    }

    if (mod) {
        p = mod->p;
        DX(4, "using module %s f %p p %p", mod->name, mod->f, mod->p);
        DX(4, "modname %s ty %d", p->name, p->type);
        c->enq = p->enqueue_wrapper;
        c->deq = p->dequeue_wrapper;
        if(p->finalprint_stats)
            c->finalprint_stats = p->finalprint_stats;
        c->si_len += p->si_datalen;
        c->q_len += p->q_datalen;
        c->schk_len += p->schk_datalen;
    }

    /* allocate queues, flowsets and one scheduler */
    c->q = calloc(c->flows, c->q_len);
    c->fs = calloc(c->flowsets, sizeof(struct dn_fsk));
    c->si = calloc(1, c->si_len);
    c->sched = calloc(c->flows, c->schk_len);

    if (c->q == NULL || c->fs == NULL) {
        D("error allocating memory for flows");
        exit(1);
    }

    /* 
     * allocate the structure and pass the 2 parameter (args and n_args)
     * for configure at runtime the double (or single) scheduler 
     */
    c->si->sched = calloc(1, sizeof(struct dn_schk));
    c->si->sched->fs = calloc(1, sizeof(struct dn_fsk));
    c->si->sched->fs->fs.args = arg_double;
    c->si->sched->fs->fs.n_args = n_arg_double;
    c->si->sched->fs->fs.qos_flows = c->flows;
    if (p) {
        if (p->config)
            p->config(c->sched);
        if (p->new_sched)
            p->new_sched(c->si);
    }
    /* parse_flowsets links queues to their flowsets */
    parse_flowsets(c, av[1], 1);
    /* complete the work calling new_fsk */
    for (i = 0; i < c->flowsets; i++) {
        if (c->fs[i].fs.par[1] == 0)
            c->fs[i].fs.par[1] = 1000;  /* default pkt len */
        c->fs[i].sched = c->sched;
        if (p && p->new_fsk)
            p->new_fsk(&c->fs[i]);
    }

    /* initialize the lists for the generator, and put
     * all flows in the list for backlog = 0
     */
    for (i=0; i <= BACKLOG+5; i++)
        INIT_LIST_HEAD(&c->ll[i]);

    for (i = 0; i < c->flows; i++) {
        struct dn_queue *q = FI2Q(c, i);
        if (q->fs == NULL)
            q->fs = &c->fs[0]; /* XXX */
        q->_si = c->si;
        

        INIT_LIST_HEAD(&q->ni.h);
        list_add_tail(&q->ni.h, &c->ll[0]);
    }
    c->llmask = 1;
    DX(5, "finished");

    return 0;
}


int
main(int ac, char *av[])
{
    struct cfg_s c;
    struct timeval end;
    double ll;
    int i;
    char msg[40];

struct timeval t1;
gettimeofday(&t1, NULL);
srand(t1.tv_usec * t1.tv_sec);
    //srand(time(NULL));
    //srand(0);
    bzero(&c, sizeof(c));
    c.ac = ac;
    c.av = av;
    init(&c);
    gettimeofday(&c.time, NULL);
    mainloop(&c);
    gettimeofday(&end, NULL);
    end.tv_sec -= c.time.tv_sec;
    end.tv_usec -= c.time.tv_usec;
    if (end.tv_usec < 0) {
        end.tv_usec += 1000000;
        end.tv_sec--;
    }
    c.time = end;
    ll = end.tv_sec*1000000 + end.tv_usec;
    ll *= 1000; /* convert to nanoseconds */
    ll /= c._enqueue;
    sprintf(msg, "1::%d", c.flows);
    DX(1,"%-8s n %d %d time %d.%06d %8.3f qlen %d %d flows %s drops %d",
        c.name, c._enqueue, c.loops,
        (int)c.time.tv_sec, (int)c.time.tv_usec, ll,
        c.th_min, c.th_max,
        c.fs_config ? c.fs_config : msg, c.drop);


    dump(&c);
    DX(2, "done ac %d av %p", ac, av);
    for (i=0; i < ac; i++)
        DX(20, "arg %d %s", i, av[i]);

    if (c.finalprint_stats)
        c.finalprint_stats(c.si);

    if (c.print_stats == 0){
        D("%-8s n %d %d time %d.%06d %8.3f qlen %d %d flows %s drops %d",
        c.name, c._enqueue, c.loops,
        (int)c.time.tv_sec, (int)c.time.tv_usec, ll,
        c.th_min, c.th_max,
        c.fs_config ? c.fs_config : msg, c.drop);
    }
    if (c.print_stats > 1) /* Finally, print execution time */
	printf("  %8.3f \n",ll);

    return 0;
}

/*
 * The controller decides whether in this iteration we should send
 * (the packet is in c->tosend) and/or receive (flag c->can_dequeue)
 */
static void
controller(struct cfg_s *c)
{
    struct dn_fs *fs;
    int flow_id;
    int i, new_i;
    struct dn_queue *q;
    struct list_head *h;

    c->to_enqueue = -1;


    /* histeresis between max and min */
    if (c->state == 0 && c->pending >= c->th_max && c->burst_left == 0)
        c->state = 1; // switch to dequeue mode
    else if (c->state == 1 && (c->pending <= c->th_min ) )
        c->state = 0; // switch to enqueue mode

    DX(3, "state %d pending %2d", c->state, c->pending);
    c->can_dequeue = c->state;
    if (c->state) /* dequeue */
        return;
    

    DX(3, "mask %x", c->llmask);

    new_i = i = ffs(c->llmask) - 1;
    if (i < 0) {
        DX(2, "no candidate (probably due to full backlogged flows)");
        c->can_dequeue = 1;
        return;
    }
    h = &c->ll[i];
    DX(1, "backlog-slot %d p %p prev %p next %p", i, h, h->prev, h->next);
    if (list_empty(h)) {
        D("failure: backlog %d empty", i);
        exit(1);
    }

    q = list_first_entry(h, struct dn_queue, ni.h);
    flow_id = Q2FI(c, q);

        DX(1, "State %d, flow_id %d, Len %d burst_left %d pendind %d, i %d"
           ,c->state, flow_id, q->ni.length, c->burst_left, c->pending, i );

#if 0
        if(i != q->ni.length)
                D("DEB %d %d\n",i,q->ni.length);
        if(i > 30)
                D("REB %d %d\n",i,q->ni.length);
#endif

    if (c->maxburst > 0 && c->burst_left == 0){
        c->burst_left = rand() % (q->fs->fs.par[0] * c->maxburst)/3 + 1; 
        if(c->burst_left + q->ni.length > BACKLOG)
            c->burst_left =  BACKLOG - q->ni.length ;
        if(c->burst_left < 0)
            c->burst_left = 0;
    }

    /* while burst is active burst_left > 0 we cannot switch to dequeue mode */
    ND("burst_left %d flow_id %ld", c->burst_left,  flow_id);
    if (c->burst_left > 1) {
        c->burst_left--;
        /* The queue artificially remains at the head of this
         * level, even if its backlog is about to grow. By
         * doing so, the flow will be the next to get a new
         * packet. This emulates a bursty arrival, and, of
         * course, happens until the burst is finished.
         */
    } else {
        list_del(&q->ni.h);
        ND("extracted flow %p flow_id %d i %d", q, flow_id, i);
        /*
         * Immediately move flow to the next level, assuming that a
         * new packet will be actually enqueued. Should the enqueue
         * fail, the drop function will be invoked and the flow will
         * be moved again to its old level.
         */
        if (list_empty(h)) {
            ND("backlog-slot %d empty", i);
            c->llmask &= ~(1<<i);
        }

        new_i = q->ni.length;
        
        if(new_i < i ||
           (c->maxburst == 0 && new_i > i)) {
            D("failure %d %d", new_i, i);
            D("new_i = %d, i = %d, c->maxburst = %d", new_i, i, c->maxburst);
            exit(1);
        }
        if (new_i < BACKLOG) {
            ND("backlog %d full", new_i+1);
            c->llmask |= 1<<(1+new_i);
            list_add_tail(&q->ni.h, h + new_i - i + 1);
        }
        c->burst_left = 0; /* burst finished */
    }
    /* 
     * the following instructions seem useless, as the fields
     * that they update are never used
     */
    fs = &q->fs->fs;
    c->cur_fs = q->fs - c->fs;
    fs->cur = flow_id;
    

    /* define the flow to be filled */
    c->to_enqueue = flow_id;


    DX(2,"y %6d flow %5d fs %3d weight %4d lenght %d",
        c->cur_y, flow_id, c->cur_fs,
        fs->par[0], q->ni.length);
}

/*
Packet allocation:
to achieve a distribution that matches weights, for each X=w/lmax class
we should generate a number of packets proportional to Y = X times the number
of flows in the class.
So we construct an array with the cumulative distribution of Y's,
and use it to identify the flow via inverse mapping (if the Y's are
not too many we can use an array for the lookup). In practice,
each flow will have X entries [virtually] pointing to it.
*/
